-- Imports
orb = module.internal('orb')
ts = module.internal('TS')
pred = module.internal('pred')
damage = module.internal('damage')

-- Menu
local mymenu = menu("GG-AIO", "GG Lucian")
---- Combo
mymenu:menu("combo", "Combo")
mymenu.combo:keybind("changerotation", "Change Combo", nil, 'J', false)
mymenu.combo:keybind("wrcombo","Semi auto W+R", 'Z', nil, false)
---- Harass
mymenu:menu("harass","Harass")
mymenu.harass:keybind("savedash", "Save dash", nil, 'U', false)
mymenu.harass:boolean("autoharass", "Auto q minion harass", true)
---- Clear
mymenu:menu("clear","Clear")
mymenu.clear:header("laneclear", "Lane Clear")
mymenu.clear:boolean("qlaneclear", "Use Q in lane clear", true)
mymenu.clear:slider("qlaneclearmana", "Minimum mana percentage needed", 30, 0, 100, 1)
mymenu.clear:slider("qminminion", "Minimum minion hit", 3, 1, 7, 1)
mymenu.clear:header("jgclear", "Jungle Clear")
mymenu.clear:boolean("qjgclear", "Use Q in jungle clear", true)
mymenu.clear:slider("qjgclearmana", "Minimum mana percentage needed", 30, 0, 100, 1)
mymenu.clear:boolean("wjgclear", "Use W in jungle clear", true)
mymenu.clear:slider("wjgclearmana", "Minimum mana percentage needed", 30, 0, 100, 1)
mymenu.clear:boolean("ejgclear", "Use E in jungle clear", true)
mymenu.clear:slider("ejgclearmana", "Minimum mana percentage needed", 30, 0, 100, 1)

---- Misc
mymenu:menu("misc","Misc")
mymenu.misc:dropdown("dash","E range",{"Short","To mouse"}) -- to add : prevent dashing out of aa range
mymenu.misc:boolean("nodashundertower", "Prevent dash under tower", true)

---- Drawings
mymenu:menu("drawings","Drawings")
mymenu.drawings:boolean("drawqext","Draw extended q range", true)
mymenu.drawings:boolean("draww","Draw w range",false)
mymenu.drawings:boolean("drawe","Draw e range",false)
mymenu.drawings:boolean("drawr","Draw r range",true)


-- Globals
local init_val = false
local mode = 1 -- both these globals are used for the combo toggle
local qrange = 500
local extendedqrange = 1000
local qwidth = 120
local wrange = 900
local maxerange = 450
local minerange = 200
local rrange = 1200

-- TODO : add W kill steal if possible (like chasing out of aa range)
-- TODO : add smart E (for example long dash if chasing (about to get out of aa range))
-- TODO : add option don't dash if about to go out of aa range	
-- TODO : add R stick to target toggle


-- Helpers
---- Cast Helpers
local function castQ(target)
	player:castSpell("obj", SpellSlot.Q, target)
end

local function castW(target,hitchance)
	local output = pred.get_prediction(target, SpellSlot.W) -- we cast it regardless of hitchance in order not lose dps for example if behind cs
	if hitchance then
		if output.hitchance >= hitchance then
			player:castSpell('pos',SpellSlot.W, output.endPos)
		end
	else
		player:castSpell('pos',SpellSlot.W, output.endPos)
	end
end

local function castE()
	local mousepos_vector = game.mousePos - player.pos
	local mousepos_vector_norm = mousepos_vector:norm()
	local dash_pos = 0
	if mymenu.misc.dash:get()==1 then
		-- doens't work if the vector is too small, min range 200
		dash_pos = player.pos + mousepos_vector_norm*minerange
	else
		if mousepos_vector:dist() > maxerange then
			dash_pos = player.pos + mousepos_vector_norm*maxerange
		elseif mousepos_vector:dist() < minerange then
			dash_pos = player.pos + mousepos_vector_norm*minerange
		else
			dash_pos = player.pos + mousepos_vector
		end
	end
	if mymenu.misc.nodashundertower:get() then
		if (not player.pos:isUnderEnemyTurret()) and (not dash_pos:isUnderEnemyTurret()) then
			player:castSpell("pos", SpellSlot.E, dash_pos)	
		end
	else
		player:castSpell("pos", SpellSlot.E, dash_pos)
	end
end

local function castR(target)
	player:castSpell('pos', SpellSlot.R, target.pos)
end

local function cast_after_attack(slot,target)
	if orb.core.can_spell_after_attack(0,true) then
		if slot == SpellSlot.Q then
			castQ(target)
		elseif slot == SpellSlot.W then
			castW(target,nil)
		elseif slot == SpellSlot.E then
			castE()
		end
	end
end

local function is_spell_ready(slot)
	return (player:spellState(slot) == SpellState.Ready)
end

local function rotation(target,spellorder)
    -- Performs the specified spell rotation using cast_after_attack
	slot1 = spellorder[1]
	slot2 = spellorder[2]
	slot3 = spellorder[3]

	if is_spell_ready(slot1) then
		cast_after_attack(slot1,target)

	elseif is_spell_ready(slot2) then
		cast_after_attack(slot2,target)

	elseif is_spell_ready(slot3) then
		cast_after_attack(slot3,target)
	end
end
---- Math Helpers
local function insidePolygon(polygon, point)
	local oddNodes = false
	local j = #polygon
	for i = 1, #polygon do
		if (polygon[i].y < point.y and polygon[j].y >= point.y or polygon[j].y < point.y and polygon[i].y >= point.y) then
			if (polygon[i].x + ( point.y - polygon[i].y ) / (polygon[j].y - polygon[i].y) * (polygon[j].x - polygon[i].x) < point.x) then
				oddNodes = not oddNodes;
			end
		end
		j = i;
	end
	return oddNodes 
end

local function compute_Q_polygon(minion)
    pm_vec = minion.pos2D - player.pos2D -- player - minion vec
    k = extendedqrange / pm_vec:len()
    pmax_vec = k*pm_vec -- pmvec resized to 1000 (extended q range)
    orth = vec2(-1*pmax_vec.y,pmax_vec.x)*(qwidth/2)/extendedqrange -- orthogonal vector for computin the rectangle, width = 60*2
    pointmax = pmax_vec + player.pos2D
    A = pointmax - orth
    B = pointmax + orth
    C = player.pos2D + orth
    D = player.pos2D - orth
    return {A,B,C,D}
end

--Combo
local function Combo()
	target = orb.core.get_target()
	if mode==1 then
		rotation(target,{SpellSlot.Q,SpellSlot.E,SpellSlot.W})
	elseif mode==2 then
		rotation(target,{SpellSlot.W,SpellSlot.E,SpellSlot.Q})
	elseif mode==3 then
		rotation(target,{SpellSlot.E,SpellSlot.W,SpellSlot.Q})
	end
end

local function wrCombo()
	player:move(game.mousePos)
	if is_spell_ready(SpellSlot.R) and not player:findBuffByName("LucianR") then
		local table = ts.get_sorted_heroes()
		local maxval = 0
		local maxobj = nil
		for i,target in pairs(table) do
			if target.value > maxval and target.obj:dist()<rrange then
				maxval = target.value
				maxobj = target.obj
			end
		end
		if maxobj and maxobj:dist()<rrange then
			castW(maxobj,nil)
			castR(maxobj)
		end
	end
end

--Harass
---- Auto harass
local function minionHarass()
    -- Long distance harass using minion
    for hero in objManager.heroes{ team = TEAM_ENEMY, dist = extendedqrange, valid_target = true } do
        -- Compute the Q cast time
        local cast_time = 0.409 - 0.009 * player.levelRef
        -- Predicted position when the Q will hit
        local future_pos = pred.core.get_pos_after_time(hero,cast_time)
        -- Check if future_pos is withing range for each minion
        for minion in objManager.minions{ team = TEAM_ENEMY, dist = qrange, valid_target = true} do
            local poly = compute_Q_polygon(minion)
            if insidePolygon(poly,future_pos) then
                castQ(minion)
            end
        end
    end
end
--
local function Harass()
	local target = orb.core.get_target()
	if target ~= nil and target.isHero then
		if mymenu.harass.savedash:get()==true then
			rotation(target,{SpellSlot.Q,SpellSlot.Q,SpellSlot.Q}) -- Very lazy
		else
			rotation(target,{SpellSlot.Q,SpellSlot.E,SpellSlot.E})
		end
	end
	minionHarass() -- if in harass mode, we do it regardless (in part because that way it can be used under tower even in auto mod, kinda lazy way)
end

-- Laneclear
local function LaneClear()
	if orb.farm.is_spell_clear_active() then
		local playerManaPercentage = player.mana/player.maxMana * 100
		if playerManaPercentage > mymenu.clear.qlaneclearmana:get() then
			-- for each minion in q range, we are looking for which one will hit the most cs
			local max_minion = nil
			local max_nb = 0
			for minion in objManager.minions{team = TEAM_ENEMY, dist = qrange, valid_target = true} do
				-- check how many minions will be hit by the q
				local poly = compute_Q_polygon(minion) -- Q hitbox
				local nb_of_minions = 0
				if not max_minion then
					max_minion = minion
				end
				-- loop over every minion in extended q range and check if within hitbox
				for minion2 in objManager.minions{team = TEAM_ENEMY, dist = extendedqrange, valid_target = true} do
					if insidePolygon(poly,minion2.pos2D) then
						nb_of_minions = nb_of_minions + 1
					end
				end
				if nb_of_minions > max_nb then
					max_nb = nb_of_minions
					max_minion = minion
				end
			end
			if max_nb >= mymenu.clear.qminminion:get() then
				castQ(max_minion)
			end
		end
	end
end

local function JungleClear()
	if orb.farm.is_spell_clear_active() then
		local playerManaPercentage = player.mana/player.maxMana * 100
		local target = orb.core.get_target()
		if target and (target.isMonster or target.isLargeMonster) then
			-- QEW logic for jungle clear
			if mymenu.clear.qjgclear and playerManaPercentage > mymenu.clear.qjgclearmana:get() and is_spell_ready(SpellSlot.Q) then
				cast_after_attack(SpellSlot.Q,target)
		
			elseif mymenu.clear.ejgclear and playerManaPercentage > mymenu.clear.ejgclearmana:get() and is_spell_ready(SpellSlot.E) then
				cast_after_attack(SpellSlot.E,target)
		
			elseif mymenu.clear.wjgclear and playerManaPercentage > mymenu.clear.wjgclearmana:get() and is_spell_ready(SpellSlot.W) then
				cast_after_attack(SpellSlot.W,target)
			end	
		end
	end
end

-- LastHit
local function LastHit()
end

-- Flee
local function Flee()
end

-- Draw
---- Hud
local hud = function()
	-- combo hud
	if mode==1 then
		rot_str = "QEW"
	elseif mode == 2 then
		rot_str = "WEQ"
	else
		rot_str = "EWQ"
	end
	graphics.draw_text_2D(rot_str, 25, player.barPos.x, player.barPos.y+170, 0xff99ff99)

    -- harass hud
	if mymenu.harass.savedash:get()==true then
		graphics.draw_text_2D("save dash", 25, player.barPos.x, player.barPos.y+190, 0xff99ff99)
	else
		graphics.draw_text_2D("save dash", 25, player.barPos.x, player.barPos.y+190, 0xffff9999)
	end
end
---- Range
local function drawRanges()
	if player.dead then return end

	if mymenu.drawings.drawqext:get() then
		graphics.draw_circle(player, extendedqrange, -1, 0xffff0000)
	end

	if mymenu.drawings.draww:get() then
		graphics.draw_circle(player, wrange, -1, 0xffff8c00)
	end

	if mymenu.drawings.drawe:get() then
		graphics.draw_circle(player, maxerange, -1, 0xff228b22)
	end

	if mymenu.drawings.drawr:get() then
		graphics.draw_circle(player, rrange, -1, 0xff0000ff)
	end
end
-- Callbacks
local function checkMode()
	if orb.core.mode() == OrbwalkingMode.Combo then
		Combo()
	elseif orb.core.mode() == OrbwalkingMode.Mixed then
    	Harass()
	elseif orb.core.mode() == OrbwalkingMode.LaneClear then
		LaneClear()
		JungleClear()
	elseif orb.core.mode() == OrbwalkingMode.LastHit then
		LastHit()
	elseif orb.core.mode() == OrbwalkingMode.Flee then
		Flee()
	end
end

local function rot_mode()
    -- Changes the combo on changerotation toggle using the global "mode"
	local cur_val = mymenu.combo.changerotation:get()
	if (cur_val ~= init_val) then
		init_val = cur_val
		mode = mode + 1
		if mode == 4 then
			mode = 1
		end
	end
end

local function autoharass()
	if mymenu.harass.autoharass:get() and orb.core.mode() ~= OrbwalkingMode.Combo then -- auto harass is disabled during combo
		if not player.pos:isUnderEnemyTurret() then
			minionHarass()
		end
	end
end

local function wrcallback()
	if mymenu.combo.wrcombo:get() then
		wrCombo()
	end
end
----
cb.add(cb.tick, wrcallback)
cb.add(cb.tick, checkMode)
cb.add(cb.tick, rot_mode)
cb.add(cb.draw_hud, hud)
cb.add(cb.tick,autoharass)
cb.add(cb.draw, drawRanges)